// aoc_proc.h: interface for the AOCProcData class.
//
//////////////////////////////////////////////////////////////////////
#ifndef AOC_PROC_H
#define AOC_PROC_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <qvariant.h>
#include "aoc_db.h"

#ifdef WIN32
#define aocDebug	qDebug
#else
#ifdef _DEBUG
#define aocDebug(f...)	qDebug(f)
#else
#define aocDebug(f...)
#endif /* _DEBUG */
#endif /* WIN32 */

typedef struct
{
	Q_UINT32 AOC_ID;
	double Multiplier;
	Q_UINT16 Value;
	Q_UINT8 Count;	// For item summary...
} AOCReq;

typedef QValueList<AOCReq> ReqsList;

typedef struct
{
	Q_UINT32 AOC_ID;
	QString Name;
	Q_UINT16 MinQL;
	Q_UINT16 MaxQL;
	Q_UINT32 AOdbID;
} AOCItem;

typedef QValueList<AOCItem> ItemList;

typedef struct
{
	AOCReq Item1;
	AOCReq Item2;
	AOCReq Product;
	Q_UINT8 Action;
	QString Note;
	ReqsList Skills;
	Q_UINT8 Count;
} AOCStep;

typedef QValueList<AOCStep> StepList;

typedef struct
{
	QString Name;
	Q_UINT16 MaxQL;
	Q_UINT16 MinQL;
	StepList Steps;
	ItemList Items;
} AOCProc;

typedef QValueList<AOCProc> ProcList;

typedef struct
{
	QString Name;
	ProcList Procs;
} AOCProcGrp;

typedef QValueList<AOCProcGrp> ProcGrpList;

class AOCProcData
{
public:
	AOCProcData();
	virtual ~AOCProcData();

	void reloadData();
	void saveData();

	void sortGroupList();
	void sortProcessList();

	enum AddMethods{ Append = -1, Prepend = -2, Replace = -3, Insert = -4 };

	int getItems();
	ItemList getItemList();
	AOCItem getItem( int item );
	AOCItem getItemByID( Q_UINT32 ID, int proc = CurrentProcess, int grp = CurrentGroup );
	int getItemIndexByID( Q_UINT32 ID, int proc = CurrentProcess, int grp = CurrentGroup );
	void addItem( AOCItem item, int index = -1 );
	void removeItem( int item );

	int getProcGroups();
	QString getProcGroup( int grp );
	void setProcGroup( int grp );
	void addProcGroup( AOCProcGrp group, int index = -1 );
	void removeProcGroup( int grp );

	int getProcesses( int grp = CurrentGroup );
	AOCProc getProcess( int proc, int grp = CurrentGroup );
	void setProcess( int proc );
	void addProcess( AOCProc proc, int index = -1, int grp = CurrentGroup );
	void removeProcess( int proc, int grp = CurrentGroup );

	int getProcItems( int proc = CurrentProcess, int grp = CurrentGroup );
	AOCItem getProcItem( int item, int proc = CurrentProcess, int grp = CurrentGroup);
	void addProcItem( AOCItem item, int index = -1, int proc = CurrentProcess, int grp = CurrentGroup );
	void removeProcItem( int item, int proc = CurrentProcess, int grp = CurrentGroup );

	int getSteps( int proc = CurrentProcess, int grp = CurrentGroup );
	AOCStep getStep( int step, int proc = CurrentProcess, int grp = CurrentGroup );
	void setStep( int step );
	void addStep( AOCStep step, int index = -1, int proc = CurrentProcess, int grp = CurrentGroup );
	void removeStep( int step, int proc = CurrentProcess, int grp = CurrentGroup );

	ReqsList makeUnique( ReqsList list );

protected:

	void loadData();

	AOCProc blankProc();
	AOCStep blankStep();
	AOCItem blankItem();
	AOCReq blankReq();

	static int CurrentGroup;
	static int CurrentProcess;
	static int CurrentStep;

	ProcGrpList ProcGroups;
	ItemList GlobalItems;

	Q_UINT32 nextFreeID();

	Q_UINT32 lastID;

#ifdef AOC_USE_BDB
	Db *AOCdb;
#endif
};

extern AOCProcData Procs;

#endif // AOC_PROC_H
